/* --------------------------------------------------------------------------
$Id: read_vhlhdf.c,v 1.8 2000/03/22 08:38:19 ahenja Exp $
Program:        

Description:    Used for reading an HDF5 file and building a node tree
		with the HighLevel protocol.

Author(s):      Anders Henja

Copyright:	Swedish Meteorological and Hydrological Institute, 2000

History:
$Log: read_vhlhdf.c,v $
Revision 1.8  2000/03/22 08:38:19  ahenja
Updated with spell checking and header information.

-----------------------------------------------------------------------------*/
#include "vhlhdf.h"
#include "read_vhlhdf.h"

/* ---------------------------------------
 * GROUP_ATTRIBUTE_ITERATOR
 * --------------------------------------- */
static herr_t groupAttributeIterator(hid_t loc_id, const char* name, void* op_data)
{
   HL_NodeList* nodeList=(HL_NodeList*)op_data;
   HL_Node* node;
   char tmp1[512],tmp2[512];
   strcpy(tmp1,nodeList->tmp_name);
   sprintf(tmp2,"%s/%s",nodeList->tmp_name,name);
   strcpy(nodeList->tmp_name,tmp2);
   addNode(nodeList, (node = newHL_Attribute(nodeList->tmp_name)));
   strcpy(nodeList->tmp_name,tmp1);
   return 0;
}

/* ---------------------------------------
 * DATASET_ATTRIBUTE_ITERATOR (IS THIS REALLY NEEDED, 
 * IT IS THE SAME CODE AS FOR GROUP_ATTRIBUTE_ITERATOR)
 * --------------------------------------- */
static herr_t datasetAttributeIterator(hid_t loc_id, const char* name, void* op_data)
{
   HL_NodeList* nodeList=(HL_NodeList*)op_data;
   HL_Node* node;
   char tmp1[512],tmp2[512];
   strcpy(tmp1,nodeList->tmp_name);
   sprintf(tmp2,"%s/%s",nodeList->tmp_name,name);
   strcpy(nodeList->tmp_name,tmp2);
   addNode(nodeList, (node = newHL_Attribute(nodeList->tmp_name)));
   strcpy(nodeList->tmp_name,tmp1);
   return 0;   
}

/* ---------------------------------------
 * GROUP_ITERATOR
 * --------------------------------------- */
static herr_t groupIterator(hid_t gid, const char* name, void* op_data)
{
   hid_t obj,type;
   H5G_stat_t statbuf;
   HL_NodeList* nodeList=(HL_NodeList*)op_data;
   HL_Node* node;
   char tmp1[512],tmp2[512];

   H5Gget_objinfo(gid,name,0,&statbuf);

   strcpy(tmp1,nodeList->tmp_name);

   switch(statbuf.type) {
   case H5G_GROUP:
      if((obj=H5Gopen(gid,name))>=0) {
	 sprintf(tmp2,"%s/%s",nodeList->tmp_name,name);
	 strcpy(nodeList->tmp_name,tmp2);
	 addNode(nodeList, (node = newHL_Group(nodeList->tmp_name)));
	 H5Aiterate(obj,NULL,groupAttributeIterator, nodeList);
	 H5Giterate(obj,".", NULL, groupIterator, nodeList);
	 strcpy(nodeList->tmp_name,tmp1);
	 H5Gclose(obj);
      }
      break;
   case H5G_DATASET:
      if((obj=H5Dopen(gid,name))>=0) {
	 sprintf(tmp2,"%s/%s",nodeList->tmp_name,name);
	 strcpy(nodeList->tmp_name,tmp2);
	 addNode(nodeList, (node=newHL_Dataset(nodeList->tmp_name)));
	 H5Aiterate(obj,NULL,datasetAttributeIterator,nodeList);
	 strcpy(nodeList->tmp_name,tmp1);
	 H5Dclose(obj);
      }
      break;
   case H5G_TYPE:
      sprintf(tmp2,"%s/%s",nodeList->tmp_name,name);
      strcpy(nodeList->tmp_name,tmp2);
      addNode(nodeList,(node=newHL_Datatype(nodeList->tmp_name)));
      strcpy(nodeList->tmp_name,tmp1);
      break;
   default:
      if(_debug_hdf)
	 fprintf(stderr,"Undefined type for %s\n",name);
      break;
   }
   return 0;
}

/* ---------------------------------------
 * READ_HL_NODE_LIST_FROM
 * --------------------------------------- */
HL_NodeList* readHL_NodeListFrom(const char* filename, const char* fromPath)
{
   hid_t file_id,gid;
   HL_NodeList* retv=NULL;
   HL_Node* node;

   if((file_id=openHlHdfFile(filename,"r"))<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to open file %s\n",filename);
      return NULL;
   }

   if(!(retv=newHL_NodeList())) {
      if(_debug_hdf)
	 fprintf(stderr,"Could not allocate NodeList\n");
      goto fail;
   }
   strcpy(retv->filename,filename);
   if((gid=H5Gopen(file_id,fromPath))<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to open root group\n");
      goto fail;
   }
   H5Aiterate(gid,NULL,groupAttributeIterator,retv);
   H5Gclose(gid);

   if(H5Giterate(file_id,fromPath,NULL,groupIterator,retv)<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Could not iterate over groups\n");
      goto fail;
   }

   H5Fclose(file_id);
   return retv;
fail:
   H5Fclose(file_id);
   freeHL_NodeList(retv);
   return NULL;   
}

/* ---------------------------------------
 * READ_HL_NODE_LIST
 * --------------------------------------- */
HL_NodeList* readHL_NodeList(const char* filename)
{
   return readHL_NodeListFrom(filename,".");
}

/* ---------------------------------------
 * SELECT_NODE
 * --------------------------------------- */
int selectNode(HL_NodeList* nodelist, const char* name)
{
   int i;

   if(!name) {
      if(_debug_hdf)
	fprintf(stderr,"Can not select any node when name is NULL\n");
      return 0;
   }

   for(i=0;i<nodelist->nNodes;i++) {
      if(strcmp(nodelist->nodes[i]->name,name)==0) {
	 nodelist->nodes[i]->mark = NMARK_SELECT;
	 return 1;
      }
   }

   if(_debug_hdf)
      fprintf(stderr, "Could not find any node called '%s'\n",name);
   return 0;
}

/* ---------------------------------------
 * SELECT_ALL_NODES
 * --------------------------------------- */
int selectAllNodes(HL_NodeList* nodelist)
{
   int i;

   for(i=0;i<nodelist->nNodes;i++) {
      nodelist->nodes[i]->mark = NMARK_SELECT;
   }

   return 1;
}

/* ---------------------------------------
 * FETCH_MARKED_NODES
 * --------------------------------------- */
int fetchMarkedNodes(HL_NodeList* nodelist)
{
   int i;
   hid_t file_id=-1;
   hid_t gid=-1;
   HL_Node* node;

   if(!(file_id=openHlHdfFile(nodelist->filename,"r"))<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Could not open file '%s' when fetching data\n",nodelist->filename);
      return 0;
   }

   if((gid=H5Gopen(file_id,"."))<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Could not open root group\n");
      goto fail;
   }

   for(i=0;i<nodelist->nNodes;i++) {
      if(nodelist->nodes[i]->mark == NMARK_SELECT) {
	 if(!fillNodeWithData(file_id, nodelist->nodes[i])) {
	    if(_debug_hdf)
	       fprintf(stderr,"Error occured when trying to fill node '%s'\n",nodelist->nodes[i]->name);
	    goto fail;
	 }
      }
   }

   if(file_id>=0) H5Fclose(file_id);
   if(gid>=0) H5Gclose(gid);
   return 1;
 fail:
   if(file_id>=0) H5Fclose(file_id);
   if(gid>=0) H5Gclose(gid);
   return 0;
}

/* ---------------------------------------
 * FILL_ATTRIBUTE_NODE
 * --------------------------------------- */
int fillAttributeNode(hid_t file_id, HL_Node* node)
{
   hid_t obj=-1;
   hid_t loc_id=-1;
   hid_t type=-1, mtype=-1;
   hid_t f_space=-1;
   size_t size;
   hsize_t* all_dims=NULL;
   hsize_t npoints;
   int ndims,i;
   HL_Node* pNode=NULL;
   char parent[512],child[512];
   unsigned char* dataptr=NULL;
   char* tmpName=NULL;
   HL_Type parentType=UNDEFINED_ID;
   H5G_stat_t statbuf;

   if(!extractParentChildName(node,parent,child)) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to extract parent/child\n");
      return 0;
   }

   if(strcmp(parent,"")!=0) {
      disableErrorReporting(); /*Just quick hack to bypass the error reporting,
				*if failed to open a dataset/or group*/
      if((loc_id=H5Gopen(file_id,parent))<0) {
	 if((loc_id=H5Dopen(file_id,parent))>=0) {
	    parentType=DATASET_ID;
	 }
      } else {
	 parentType=GROUP_ID;
      }
      enableErrorReporting();
      if(loc_id<0) {
	 if(_debug_hdf)
	    fprintf(stderr,"Parent '%s' to attribute '%s' could not be opened\n",
		    parent,child);
	 return 0;
      }      
   } else {
      if((loc_id=H5Gopen(file_id,"/"))<0) {
	 if(_debug_hdf)
	    fprintf(stderr,"Could not open root group when reading attr '%s'\n",
		    child);
	 return 0;
      }
      parentType=GROUP_ID;
   }

   if((obj=H5Aopen_name(loc_id,child))<0) {
      return 0;
   }

   if((type=H5Aget_type(obj))<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Could not get attribute type\n");
      goto fail;
   }
   if((mtype=getFixedType(type))<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Could not create fixed attribute type\n");
      goto fail;
   }
   if(H5Tget_class(mtype)==H5T_COMPOUND) {
      if(H5Tcommitted(type)>0) {
	 H5Gget_objinfo(type,".",TRUE,&statbuf);
      }
      if(!(node->compoundDescription = buildTypeDescriptionFromTypeHid(mtype))) {
	 if(_debug_hdf)
	    fprintf(stderr,"Failed to create compound data description for attribute\n");
	 goto fail;
      }
      if(H5Tcommitted(type)>0) {
	 node->compoundDescription->objno[0]=statbuf.objno[0];
	 node->compoundDescription->objno[1]=statbuf.objno[1];
      }
   }
   size=H5Tget_size(type);
   node->dSize=size;
   if((f_space=H5Aget_space(obj))>=0) {
      ndims = H5Sget_simple_extent_ndims(f_space);
      npoints = H5Sget_simple_extent_npoints(f_space);
      if(ndims>0) { 
	 if(!(all_dims=malloc(sizeof(hsize_t)*ndims))) {
	    if(_debug_hdf)
	       fprintf(stderr,"Could not allocate memory for node dims\n");
	    goto fail;
	 }
      }
#ifdef STUPID_BUG_BY_ME_AHE
      /* If scalar, ndims==0, but it still will be correct*/
      printf("NDIMS=%d, NPOINTS=%d, SIZE=%d\n",ndims,npoints,size);
      if(!(all_dims=malloc(sizeof(hsize_t)*ndims))) {
	 if(_debug_hdf)
	    fprintf(stderr,"Could not allocate memory for node dims\n");
	 goto fail;
      }
#endif
      H5Sget_simple_extent_dims(f_space,all_dims,NULL);
      if(H5Sis_simple(f_space)>=0) {
	 if(!(dataptr=malloc(H5Tget_size(mtype)*npoints))) {
	    if(_debug_hdf)
	       fprintf(stderr,"Could not allocate memory for attribute data\n");
	    goto fail;
	 }
	 if(H5Aread(obj,mtype,dataptr)<0) {
	    if(_debug_hdf)
	       fprintf(stderr,"Could not read attribute data\n");
	    goto fail;
	 }
      } else {
	 if(_debug_hdf)
	    fprintf(stderr,"Attribute dataspace was not simple, can't handle\n");
	 goto fail;
      }
   } else {
      if(_debug_hdf)
	 fprintf(stderr,"Could not get dataspace for attribute\n");
      goto fail;
   }

   if(!(tmpName=getFormatNameString(mtype))) {
      goto fail;
   }
   strcpy(node->format,tmpName);

   if(all_dims) {
      for(i=0;i<ndims;i++)
	 node->dims[i]=all_dims[i];
      free(all_dims);
   }
   node->ndims=ndims;
   node->typeId=mtype;

   if(obj>=0) H5Aclose(obj);
   if(loc_id>=0) {
      if(parentType==GROUP_ID) {
	 H5Gclose(loc_id);
      } else if(parentType==DATASET_ID) {
	 H5Dclose(loc_id);
      } else {
	 if(_debug_hdf)
	    fprintf(stderr,"Could not determine type of loc_id for '%s', could not close\n",
		    parent);
      }
   }
   /* Mark the node as original */
   node->mark = NMARK_ORIGINAL;

   if(type>=0) H5Tclose(type);
   if(f_space>=0) H5Sclose(f_space);
   node->data=dataptr;
   if(tmpName) free(tmpName);
   return 1;
fail:
   if(obj>=0) H5Aclose(obj);
   if(loc_id>=0) {
      if(parentType==GROUP_ID) {
	 H5Gclose(loc_id);
      } else if(parentType==DATASET_ID) {
	 H5Dclose(loc_id);
      } else {
	 if(_debug_hdf)
	    fprintf(stderr,"Could not determine type of loc_id, could not close\n");
      }
   }
   if(type>=0) H5Tclose(type);
   if(mtype>=0) H5Tclose(mtype);
   if(f_space>=0) H5Sclose(f_space);
   if(all_dims) free(all_dims);
   if(dataptr) free(dataptr);
   if(tmpName) free(tmpName);

   return 0;
}

/* ---------------------------------------
 * FILL_DATASET_NODE
 * --------------------------------------- */
int fillDatasetNode(hid_t file_id, HL_Node* node)
{
   hid_t obj=-1;
   hid_t type=-1;
   H5G_stat_t statbuf;
   size_t size;
   int ndims,i;
   hsize_t* all_dims=NULL;
   hsize_t npoints;
   hid_t f_space=-1;
   hid_t mtype=-1;
   hid_t databuff_space=-1;
   unsigned char* dataptr=NULL;
   char* tmpName=NULL;

   if((obj=H5Dopen(file_id,node->name))<0) {
      goto fail;
   }
   /* What datatype was this dataset stored as? */
   if((type=H5Dget_type(obj))<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to get type from dataset\n");
      goto fail;
   }
   

   /* What size does the type have? */
   size=H5Tget_size(type);
   node->dSize=size;
   if((f_space=H5Dget_space(obj))>0) { /*Get the space description for the dataset */
      ndims=H5Sget_simple_extent_ndims(f_space);
      npoints=H5Sget_simple_extent_npoints(f_space);
      all_dims=malloc(sizeof(hsize_t)*ndims);
      if(H5Sget_simple_extent_dims(f_space,all_dims,NULL)!=ndims) {
	 if(_debug_hdf)
	    fprintf(stderr,"Could not get dimensions from dataset\n");
      }
      /* Translate the type into a native dataspace */
      mtype = getFixedType(type);
      if(H5Tget_class(mtype)==H5T_COMPOUND) {
	 if(H5Tcommitted(type)>0) {
	    H5Gget_objinfo(type,".",TRUE,&statbuf);
	 }
	 if(!(node->compoundDescription = buildTypeDescriptionFromTypeHid(mtype))) {
	    if(_debug_hdf)
	       fprintf(stderr,"Failed to create compound data description for dataset\n");
	    goto fail;
	 }
	 if(H5Tcommitted(type)>0) {
	    node->compoundDescription->objno[0]=statbuf.objno[0];
	    node->compoundDescription->objno[1]=statbuf.objno[1];
	 }
      }
      if(H5Sis_simple(f_space)>=0) { /*Only allow simple dataspace, nothing else supported by HDF5 anyway */
	 dataptr=malloc(H5Tget_size(mtype)*npoints);
	 databuff_space=H5Screate_simple(1,&npoints,NULL);
	 H5Sselect_all(f_space); /* Mark for selection */
	 H5Sselect_all(databuff_space);  /* Mark for selection */
	 if(H5Dread(obj,mtype,databuff_space,f_space,H5P_DEFAULT,dataptr)<0) {
	    if(_debug_hdf)
	       fprintf(stderr,"Failed to read dataset\n");
	    goto fail;
	 }
      } else {
	 if(_debug_hdf)
	    fprintf(stderr,"Dataspace for dataset was not simple, this is not supported\n");
	 goto fail;
      }
   }

   /* Translate the type into a recognizable string, e.g. int, short, string, compound */
   if(!(tmpName=getFormatNameString(mtype))) {
      goto fail;
   }
   strcpy(node->format,tmpName);

   /* Filling dataset */
   node->ndims=ndims;
   for(i=0;i<ndims;i++) {
      node->dims[i]=all_dims[i];
   }

   /* Remember to keep the type pointer, can be hard to interpreet the data
    * later otherwise. */
   node->typeId=mtype;

   /*Better keep data also*/
   node->data=dataptr;

   /* Mark the node as original */
   node->mark = NMARK_ORIGINAL;

   if(obj>=0) H5Dclose(obj);
   if(type>=0) H5Tclose(type);
   if(all_dims) free(all_dims);
   if(f_space>=0) H5Sclose(f_space);
   if(databuff_space>=0) H5Sclose(databuff_space);
   if(tmpName) free(tmpName);

   return 1;
 fail:
   if(obj>=0) H5Dclose(obj);
   if(type>=0) H5Tclose(type);
   if(all_dims) free(all_dims);
   if(f_space>=0) H5Sclose(f_space);
   if(mtype>=0) H5Tclose(mtype);
   if(databuff_space>=0) H5Sclose(databuff_space);
   if(dataptr) free(dataptr);
   if(tmpName) free(tmpName);
   return 0;
}

/* ---------------------------------------
 * FILL_GROUP_NODE
 * --------------------------------------- */
int fillGroupNode(hid_t file_id, HL_Node* node)
{
   hid_t obj;

   if((obj=H5Gopen(file_id,node->name))<0) {
      return 0;
   }

   /* Mark the node as original */
   node->mark = NMARK_ORIGINAL;

   H5Gclose(obj);
   return 1;
}

HL_CompoundTypeDescription* buildTypeDescriptionFromTypeHid(hid_t type_id)
{
   HL_CompoundTypeDescription* typelist=NULL;
   HL_CompoundTypeAttribute* typenode;
   int i,j,nmembers;
   char* fname;
   hid_t mtype=-1;
   hid_t fixedType=-1;
   size_t* dims=NULL;
   size_t dSize;
   int ndims;
   char* tmpName=NULL;

   if(!(typelist=newHL_CompoundTypeDescription())) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to create datatype nodelist\n");
      goto fail;
   }

   typelist->size=H5Tget_size(type_id);
   nmembers = H5Tget_nmembers(type_id);
   for(i=0;i<nmembers;i++) {
      size_t offset = H5Tget_member_offset(type_id,i);
      fname = H5Tget_member_name(type_id,i);
      mtype = H5Tget_member_type(type_id,i);

      /*Added by AHE 000315*/
      if((fixedType = getFixedType(mtype))<0) {
	 if(_debug_hdf)
	    fprintf(stderr,"Failed to convert to fixed type\n");
	 goto fail;
      }

      if(!(tmpName=getFormatNameString(fixedType))) {
	 goto fail;
      }

      ndims = H5Tget_member_dims(type_id,i,NULL,NULL);
      dims = malloc(sizeof(size_t)*ndims);
      ndims = H5Tget_member_dims(type_id,i,dims,NULL);
      dSize = H5Tget_size(type_id);
      
      typenode=newHL_CompoundTypeAttribute(fname,offset,tmpName,dSize,ndims,dims);
      if(!addCompoundTypeAttribute(typelist,typenode))
	 goto fail;
      if(dims)
	 free(dims);
      dims=NULL;
      if(mtype>=0)
	 H5Tclose(mtype);
      mtype=-1;
      if(fname)
	 free(fname);
      fname=NULL;
      if(tmpName)
	 free(tmpName);
      tmpName=NULL;
      if(fixedType>=0)
	 H5Tclose(fixedType);
      fixedType=-1;
   }

   if(dims) free(dims);
   if(mtype>=0) H5Tclose(mtype);
   if(fixedType>=0)
      H5Tclose(fixedType);
   if(fname) free(fname);
   if(tmpName) free(tmpName);
   return typelist;
fail:
   if(dims) free(dims);
   if(mtype>=0) H5Tclose(mtype);
   if(fixedType>=0)
      H5Tclose(fixedType);
   if(fname) free(fname);
   if(tmpName) free(tmpName);
   freeHL_CompoundTypeDescription(typelist);
   return NULL;
}

/* ---------------------------------------
 * FILL_TYPE_NODE
 * --------------------------------------- */
int fillTypeNode(hid_t file_id, HL_Node* node)
{
   hid_t obj=-1;
   HL_CompoundTypeDescription* typelist=NULL;
   H5G_stat_t statbuf;
   
   if((obj=H5Topen(file_id,node->name))<0) {
      return 0;
   }

   H5Gget_objinfo(obj,".",TRUE,&statbuf);

   if(!(typelist=buildTypeDescriptionFromTypeHid(obj))) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to create datatype nodelist\n");
      goto fail;
   }
   strcpy(typelist->typename,node->name);
   typelist->objno[0]=statbuf.objno[0];
   typelist->objno[1]=statbuf.objno[1];

   node->compoundDescription=typelist;

   /* Mark the node as original */
   node->mark = NMARK_ORIGINAL;

   if(obj>=0) H5Tclose(obj);
   return 1;
fail:
   if(obj>=0) H5Tclose(obj);
   freeHL_CompoundTypeDescription(typelist);
   return 0;
}

/* ---------------------------------------
 * FILL_NODE_WITH_DATA
 * --------------------------------------- */
int fillNodeWithData(hid_t file_id, HL_Node* node)
{
   switch(node->type) {
   case ATTRIBUTE_ID:
      return fillAttributeNode(file_id, node);
      /*break;*/
   case DATASET_ID:
      return fillDatasetNode(file_id, node);
      /*break;*/
   case GROUP_ID:
      return fillGroupNode(file_id,node);
      /*break;*/
   case TYPE_ID:
      return fillTypeNode(file_id,node);
      /*break;*/
   default:
      if(_debug_hdf)
	 fprintf(stderr,"Can't handle other nodetypes but '%d'\n",node->type);
      break;
   }
   return 0;
}

